/*-
 * Copyright (c) 1990, 1993
 *  The Regents of the University of California.  All rights reserved.
 *
 * This code is derived from software contributed to Berkeley by
 * Donn Seeley at UUNET Technologies, Inc.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *  This product includes software developed by the University of
 *  California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/*******************************/
/* vsscanf.c fur Hitachi CPU   */
/*                   '01/10/26 */
/*           by BestTechnology */
/*******************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <stdarg.h>
#include "uart.h"

#ifdef __cplusplus
extern "C" {
#endif

#define vssanf_BUF			32		/* Maximum length of numeric string. */

/*
 * Flags used during conversion.
 */
#define vssanf_LONG			0x01	/* l: long or double */
#define vssanf_LONGDBL		0x02	/* L: long double; unimplemented */
#define vssanf_SHORT		0x04	/* h: short */
#define vssanf_SUPPRESS		0x08	/* suppress assignment */
#define vssanf_POINTER		0x10	/* weird %p pointer (`fake hex') */
#define vssanf_NOSKIP		0x20	/* do not skip blanks */

/*
 * The following are used in numeric conversions only:
 * SIGNOK, NDIGITS, DPTOK, and EXPOK are for floating point;
 * SIGNOK, NDIGITS, PFXOK, and NZDIGITS are for integral.
 */
#define vssanf_SIGNOK		0x40	/* +/- is (still) legal */
#define vssanf_NDIGITS		0x80	/* no digits detected */

#define vssanf_DPTOK		0x100	/* (float) decimal point is still legal */
#define vssanf_EXPOK		0x200	/* (float) exponent (e+3, etc) still legal */

#define vssanf_PFXOK		0x100	/* 0x prefix is (still) legal */
#define vssanf_NZDIGITS		0x200	/* no zero digits detected */

/*
 * Conversion types.
 */
#define CT_CHAR		0	/* %c conversion */
#define CT_CCL		1	/* %[...] conversion */
#define CT_STRING	2	/* %s conversion */
#define CT_INT		3	/* integer, i.e., strtol or strtoul */
#define CT_FLOAT	4	/* floating, i.e., strtod */

#define vssanf_u_char unsigned char
#define vssanf_u_long unsigned long

static vssanf_u_char *__sccl(register char *tab, register vssanf_u_char *fmt);

/*
 * v_sscanf
 */
int v_sscanf(char *str, char const *fmt0, va_list ap)
{
	register vssanf_u_char *fmt = (vssanf_u_char *)fmt0;
	register int c;		/* character from format, or conversion */
	register size_t width;	/* field width, or 0 */
	register char *p;	/* points into all kinds of strings */
	register int n;		/* handy integer */
	register int flags;	/* flags as defined above */
	register char *p0;	/* saves original value of p when necessary */
	register int index;	/* str[index] */
	int len;			/* str[] length */
	int nassigned;		/* number of fields assigned */
	int nread;			/* number of characters consumed from fp */
	int base;			/* base argument to strtol/strtoul */
	vssanf_u_long (*ccfn)(const char *, char **, int);	/* conversion function (strtol/strtoul) */
	char ccltab[256];	/* character class table for %[...] */
	char buf[vssanf_BUF];		/* buffer for numeric conversions */

	/* `basefix' is used to avoid `if' tests in the integer scanner */
	static short basefix[17] =
		{ 10, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16 };

	len = strlen(str);
	index = 0;
	nassigned = 0;
	nread = 0;
	base = 0;		/* XXX just to keep gcc happy */
	ccfn = NULL;		/* XXX just to keep gcc happy */
	for (;;) {
		c = *fmt++;
		if (c == 0)
			return (nassigned);
		if (isspace(c)) {
			for (;;) {
				if (len <= 0)
					return (nassigned);
				if (!isspace(str[index]))
					break;
				nread++, len--, index++;
			}
			continue;
		}
		if (c != '%')
			goto literal;
		width = 0;
		flags = 0;
		/*
		 * switch on the format.  continue if done;
		 * break once format type is derived.
		 */
again:		c = *fmt++;
		switch (c) {
		case '%':
literal:
			if (len <= 0)
				goto input_failure;
			if (str[index] != c)
				goto match_failure;
			len--, index;
			nread++;
			continue;

		case '*':
			flags |= vssanf_SUPPRESS;
			goto again;
		case 'l':
			flags |= vssanf_LONG;
			goto again;
		case 'L':
			flags |= vssanf_LONGDBL;
			goto again;
		case 'h':
			flags |= vssanf_SHORT;
			goto again;

		case '0': case '1': case '2': case '3': case '4':
		case '5': case '6': case '7': case '8': case '9':
			width = width * 10 + c - '0';
			goto again;

		/*
		 * Conversions.
		 * Those marked `compat' are for 4.[123]BSD compatibility.
		 *
		 * (According to ANSI, E and X formats are supposed
		 * to the same as e and x.  Sorry about that.)
		 */
		case 'D': /* compat */
			flags |= vssanf_LONG;
			/* FALLTHROUGH */
		case 'd':
			c = CT_INT;
			ccfn = (vssanf_u_long (*)(const char *, char **, int))strtol;
			base = 10;
			break;

		case 'i':
			c = CT_INT;
			ccfn = (vssanf_u_long (*)(const char *, char **, int))strtol;
			base = 0;
			break;

		case 'O': /* compat */
			flags |= vssanf_LONG;
			/* FALLTHROUGH */
		case 'o':
			c = CT_INT;
			ccfn = strtoul;
			base = 8;
			break;

		case 'u':
			c = CT_INT;
			ccfn = strtoul;
			base = 10;
			break;

		case 'X': /* compat   XXX */
			flags |= vssanf_LONG;
			/* FALLTHROUGH */
		case 'x':
			flags |= vssanf_PFXOK; /* enable 0x prefixing */
			c = CT_INT;
			ccfn = strtoul;
			base = 16;
			break;
		case 's':
			c = CT_STRING;
			break;
		case '[':
			fmt = __sccl(ccltab, fmt);
			flags |= vssanf_NOSKIP;
			c = CT_CCL;
			break;
		case 'c':
			flags |= vssanf_NOSKIP;
			c = CT_CHAR;
			break;
		case 'p': /* pointer format is like hex */
			flags |= vssanf_POINTER | vssanf_PFXOK;
			c = CT_INT;
			ccfn = strtoul;
			base = 16;
			break;

		case 'n':
			if (flags & vssanf_SUPPRESS) /* ??? */
				continue;
			if (flags & vssanf_SHORT)
				*va_arg(ap, short *) = nread;
			else if (flags & vssanf_LONG)
				*va_arg(ap, long *) = nread;
			else
				*va_arg(ap, int *) = nread;
			continue;

		/*
		 * Disgusting backwards compatibility hacks.  XXX
		 */
		case '\0':	/* compat */
			return (EOF);

		default:	/* compat */
			if (isupper(c))
				flags |= vssanf_LONG;
			c = CT_INT;
			ccfn = (vssanf_u_long (*)(const char *, char **, int))strtol;
			base = 10;
			break;
		}

		/*
		 * We have a conversion that requires input.
		 */
		if (len <= 0)
			goto input_failure;

		/*
		 * Consume leading white space, except for formats
		 * that suppress this.
		 */
		if ((flags & vssanf_NOSKIP) == 0) {
			while (isspace(str[index])) {
				nread++;
				if (--len > 0)
					index++;
				else
					goto input_failure;
			}
			/*
			 * Note that there is at least one character in
			 * the buffer, so conversions that do not set vssanf_NOSKIP
			 * ca no longer result in an input failure.
			 */
		}

		/*
		 * Do the conversion.
		 */
		switch (c) {

		case CT_CHAR:
			/* scan arbitrary characters (sets vssanf_NOSKIP) */
			if (width == 0)
				width = 1;
			if (flags & vssanf_SUPPRESS) {
				size_t sum = 0;
				for (;;) {
					if ((n = len) < width) {
						sum += n;
						width -= n;
						index += n;
						if (sum == 0)
								goto input_failure;
						break;
					} else {
						sum += width;
						len -= width;
						index += width;
						break;
					}
				}
				nread += sum;
			} else {
				size_t r = (len>width) ? width : len;

				if (r == 0)
					goto input_failure;
				nread += r;
				nassigned++;
			}
			break;

		case CT_CCL:
			/* scan a (nonempty) character class (sets vssanf_NOSKIP) */
			if (width == 0)
				width = (size_t)~0; /* `infinity' */
			/* take only those things in the class */
			if (flags & vssanf_SUPPRESS) {
				n = 0;
				while (ccltab[str[index]]) {
					n++, len--, index++;
					if (--width == 0)
						break;
					if (len <= 0) {
						if (n == 0)
							goto input_failure;
						break;
					}
				}
				if (n == 0)
					goto match_failure;
			} else {
				p0 = p = va_arg(ap, char *);
				while (ccltab[str[index]]) {
					len--;
					*p++ = str[index++];
					if (--width == 0)
						break;
					if (len <= 0) {
						if (p == p0)
							goto input_failure;
						break;
					}
				}
				n = p - p0;
				if (n == 0)
					goto match_failure;
				*p = 0;
				nassigned++;
			}
			nread += n;
			break;

		case CT_STRING:
			/* like CCL, but zero-length string OK, & no vssanf_NOSKIP */
			if (width == 0)
				width = (size_t)~0;
			if (flags & vssanf_SUPPRESS) {
				n = 0;
				while (!isspace(str[index])) {
					n++, len--, index++;
					if (--width == 0)
						break;
					if (len <= 0)
						break;
				}
				nread += n;
			} else {
				p0 = p = va_arg(ap, char *);
				while (!isspace(str[index])) {
					len--;
					*p++ = str[index++];
					if (--width == 0)
						break;
					if (len <= 0)
						break;
				}
				*p = 0;
				nread += p - p0;
				nassigned++;
			}
			continue;

		case CT_INT:
			/* scan an integer as if by strtol/strtoul */
#ifdef hardway
			if (width == 0 || width > sizeof(buf) - 1)
				width = sizeof(buf) - 1;
#else
			/* size_t is unsigned, hence this optimisation */
			if (--width > sizeof(buf) - 2)
				width = sizeof(buf) - 2;
			width++;
#endif
			flags |= vssanf_SIGNOK | vssanf_NDIGITS | vssanf_NZDIGITS;
			for (p = buf; width; width--) {
				c = str[index];
				/*
				 * Switch on the character; `goto ok'
				 * if we accept it as a part of number.
				 */
				switch (c) {

				/*
				 * The digit 0 is always legal, but is
				 * special.  For %i conversions, if no
				 * digits (zero or nonzero) have been
				 * scanned (only signs), we will have
				 * base==0.  In that case, we should set
				 * it to 8 and enable 0x prefixing.
				 * Also, if we have not scanned zero digits
				 * before this, do not turn off prefixing
				 * (someone else will turn it off if we
				 * have scanned any nonzero digits).
				 */
				case '0':
					if (base == 0) {
						base = 8;
						flags |= vssanf_PFXOK;
					}
					if (flags & vssanf_NZDIGITS)
							flags &= ~(vssanf_SIGNOK|vssanf_NZDIGITS|vssanf_NDIGITS);
					else
							flags &= ~(vssanf_SIGNOK|vssanf_PFXOK|vssanf_NDIGITS);
					goto ok;

				/* 1 through 7 always legal */
				case '1': case '2': case '3':
				case '4': case '5': case '6': case '7':
					base = basefix[base];
					flags &= ~(vssanf_SIGNOK | vssanf_PFXOK | vssanf_NDIGITS);
					goto ok;

				/* digits 8 and 9 ok iff decimal or hex */
				case '8': case '9':
					base = basefix[base];
					if (base <= 8)
						break;	/* not legal here */
					flags &= ~(vssanf_SIGNOK | vssanf_PFXOK | vssanf_NDIGITS);
					goto ok;

				/* letters ok iff hex */
				case 'A': case 'B': case 'C':
				case 'D': case 'E': case 'F':
				case 'a': case 'b': case 'c':
				case 'd': case 'e': case 'f':
					/* no need to fix base here */
					if (base <= 10)
						break;	/* not legal here */
					flags &= ~(vssanf_SIGNOK | vssanf_PFXOK | vssanf_NDIGITS);
					goto ok;

				/* sign ok only as first character */
				case '+': case '-':
					if (flags & vssanf_SIGNOK) {
						flags &= ~vssanf_SIGNOK;
						goto ok;
					}
					break;

				/* x ok iff flag still set & 2nd char */
				case 'x': case 'X':
					if (flags & vssanf_PFXOK && p == buf + 1) {
						base = 16;	/* if %i */
						flags &= ~vssanf_PFXOK;
						goto ok;
					}
					break;
				}

				/*
				 * If we got here, c is not a legal character
				 * for a number.  Stop accumulating digits.
				 */
				break;
		ok:
				/*
				 * c is legal: store it and look at the next.
				 */
				*p++ = c;
				if(len > 0)
					index++;
				else
					break;	/* EOF */
			}
			/*
			 * If we had only a sign, it is no good; push
			 * back the sign.  If the number ends in `x',
			 * it was [sign] '0' 'x', so push back the x
			 * and treat it as [sign] '0'.
			 */
			if (flags & vssanf_NDIGITS) {
				if (p > buf) {
					/* ungetc */
					str[--index] = *(vssanf_u_char *)--p;
					len++;
				}
				goto match_failure;
			}
			c = ((vssanf_u_char *)p)[-1];
			if (c == 'x' || c == 'X') {
				--p;
				/* ungetc */
				str[--index] = c;
				len++;
			}
			if ((flags & vssanf_SUPPRESS) == 0) {
				vssanf_u_long res;

				*p = 0;
				res = (*ccfn)(buf, (char **)NULL, base);
				if (flags & vssanf_POINTER)
					*va_arg(ap, void **) = (void *)res;
				else if (flags & vssanf_SHORT)
					*va_arg(ap, short *) = res;
				else if (flags & vssanf_LONG)
					*va_arg(ap, long *) = res;
				else
					*va_arg(ap, int *) = res;
				nassigned++;
			}
			nread += p - buf;
			break;
		}
	}
input_failure:
	return (nassigned ? nassigned : -1);
match_failure:
	return (nassigned);
}

/*
 * Fill in the given table from the scanset at the given format
 * (just after `[').  Return a pointer to the character past the
 * closing `]'.  The table has a 1 wherever characters should be
 * considered part of the scanset.
 */
static vssanf_u_char *
__sccl(register char *tab, register vssanf_u_char *fmt)
{
	register int c, n, v;

	/* first `clear' the whole table */
	c = *fmt++;		/* first char hat => negated scanset */
	if (c == '^') {
		v = 1;		/* default => accept */
		c = *fmt++; /* get new first char */
	} else
		v = 0;		/* default => reject */
	/* should probably use memset here */
	for (n = 0; n < 256; n++)
		tab[n] = v;
	if (c == 0)
		return (fmt - 1);/* format ended before closing ] */

	/*
	 * Now set the entries corresponding to the actual scanset
	 * to the opposite of the above.
	 *
	 * The first character may be ']' (or '-') without being special;
	 * the last character may be '-'.
	 */
	v = 1 - v;
	for (;;) {
		tab[c] = v;		/* take character c */
doswitch:
		n = *fmt++;		/* and examine the next */
		switch (n) {

		case 0:			/* format ended too soon */
			return (fmt - 1);

		case '-':
			/*
			 * A scanset of the form
			 *  [01+-]
			 * is defined as `the digit 0, the digit 1,
			 * the character +, the character -', but
			 * the effect of a scanset such as
			 *  [a-zA-Z0-9]
			 * is implementation defined.  The V7 Unix
			 * scanf treats `a-z' as `the letters a through
			 * z', but treats `a-a' as `the letter a, the
			 * character -, and the letter a'.
			 *
			 * For compatibility, the `-' is not considerd
			 * to define a range if the character following
			 * it is either a close bracket (required by ANSI)
			 * or is not numerically greater than the character
			 * we just stored in the table (c).
			 */
			n = *fmt;
			if (n == ']' || n < c) {
				c = '-';
				break;	/* resume the for(;;) */
			}
			fmt++;
			do {		/* fill in the range */
				tab[++c] = v;
			} while (c < n);
#if 1 /* XXX another disgusting compatibility hack */
			/*
			 * Alas, the V7 Unix scanf also treats formats
			 * such as [a-c-e] as `the letters a through e'.
			 * This too is permitted by the standard....
			 */
			goto doswitch;
#else
			c = *fmt++;
			if (c == 0)
				return (fmt - 1);
			if (c == ']')
				return (fmt);
#endif
			break;

		case ']':		/* end of scanset */
			return (fmt);

		default:		/* just another character */
			c = n;
			break;
		}
	}
	/* NOTREACHED */
}

#ifdef __cplusplus
}
#endif

int uart0_scanf(const char *fmt, ...)
{
	va_list ap;
	int ret;
	char ibuf[vssanf_BUF];

	uart0_gets(ibuf, vssanf_BUF-2);
	va_start(ap, fmt);
	ret = v_sscanf(ibuf, fmt, ap);
	va_end(ap);
	return (ret);
}

